# -*- coding: utf-8 -*-

import pytest
from pymatgen.analysis.diffusion.neb.io import MVLCINEBEndPointSet, MVLCINEBSet
from pymatgen.io.vasp.sets import MITNEBSet  # MVLNPTMDSet,
from pymatgen.io.vasp.sets import (
    MITMDSet,
    MITRelaxSet,
    MPHSEBSSet,
    MPHSERelaxSet,
    MPMDSet,
    MPMetalRelaxSet,
    MPNMRSet,
    MPNonSCFSet,
    MPRelaxSet,
    MPScanRelaxSet,
    MPScanStaticSet,
    MPStaticSet,
    MVLElasticSet,
    MVLGBSet,
)

from simmate.calculators.vasp.inputs import Incar, Potcar
from simmate.calculators.vasp.workflows import (  # Dynamics__Vasp__MvlNpt,
    Diffusion__Vasp__NebFromImagesMit,
    Diffusion__Vasp__NebFromImagesMvlCi,
    Dynamics__Vasp__Matproj,
    Dynamics__Vasp__Mit,
    ElectronicStructure__Vasp__MatprojBandStructure,
    ElectronicStructure__Vasp__MatprojBandStructureHse,
    ElectronicStructure__Vasp__MatprojDensityOfStates,
    ElectronicStructure__Vasp__MatprojDensityOfStatesHse,
    Relaxation__Vasp__Matproj,
    Relaxation__Vasp__MatprojHse,
    Relaxation__Vasp__MatprojMetal,
    Relaxation__Vasp__MatprojScan,
    Relaxation__Vasp__Mit,
    Relaxation__Vasp__MvlGrainboundary,
    Relaxation__Vasp__MvlNebEndpoint,
    Relaxation__Vasp__MvlSlab,
    StaticEnergy__Vasp__Matproj,
    StaticEnergy__Vasp__MatprojScan,
)
from simmate.calculators.vasp.workflows.elastic.mvl import Elastic__Vasp__Mvl
from simmate.calculators.vasp.workflows.nuclear_magnetic_resonance import (
    Nmr__Vasp__MatprojChemicalShifts,
    Nmr__Vasp__MatprojFieldGradient,
)
from simmate.conftest import make_dummy_files
from simmate.toolkit.diffusion import MigrationImages

MD_KWARGS = {
    "start_temp": 300,
    "end_temp": 1200,
    "nsteps": 10000,
}


@pytest.mark.pymatgen
@pytest.mark.parametrize(
    "simmate_workflow, pymatgen_set, pymatgen_kwargs",
    [
        (Relaxation__Vasp__Matproj, MPRelaxSet, {}),
        (Relaxation__Vasp__Mit, MITRelaxSet, {}),
        (Relaxation__Vasp__MatprojHse, MPHSERelaxSet, {}),
        (Relaxation__Vasp__MatprojMetal, MPMetalRelaxSet, {}),
        (Relaxation__Vasp__MatprojScan, MPScanRelaxSet, {}),
        (Relaxation__Vasp__MvlNebEndpoint, MVLCINEBEndPointSet, {}),
        (Relaxation__Vasp__MvlGrainboundary, MVLGBSet, {"slab_mode": False}),
        (Relaxation__Vasp__MvlSlab, MVLGBSet, {"slab_mode": True}),
        (StaticEnergy__Vasp__Matproj, MPStaticSet, {}),
        (StaticEnergy__Vasp__MatprojScan, MPScanStaticSet, {}),
        (Dynamics__Vasp__Mit, MITMDSet, MD_KWARGS),
        (Dynamics__Vasp__Matproj, MPMDSet, MD_KWARGS),
        # MVL-NPT MD requires POTCARs to be configured for pymatgen. need a workaround.
        # (Dynamics__Vasp__MvlNpt, MVLNPTMDSet, MD_KWARGS),
        (
            ElectronicStructure__Vasp__MatprojBandStructureHse,
            MPHSEBSSet,
            {"mode": "line"},
        ),
        (
            ElectronicStructure__Vasp__MatprojBandStructure,
            MPNonSCFSet,
            {"mode": "line"},
        ),
        (
            ElectronicStructure__Vasp__MatprojDensityOfStates,
            MPNonSCFSet,
            {"mode": "uniform"},
        ),
        (
            ElectronicStructure__Vasp__MatprojDensityOfStatesHse,
            MPHSEBSSet,
            {"mode": "uniform"},
        ),
        (Nmr__Vasp__MatprojChemicalShifts, MPNMRSet, {"mode": "cs"}),
        (Nmr__Vasp__MatprojFieldGradient, MPNMRSet, {"mode": "efg"}),
        (Elastic__Vasp__Mvl, MVLElasticSet, {}),
    ],
)
def test_pymatgen_input_sets(
    structure,
    tmp_path,
    mocker,
    simmate_workflow,
    pymatgen_set,
    pymatgen_kwargs,
):
    """
    Many of the presets implemented are ported directly from pymatgen, so these
    tests simply confirm that the inputs give the same INCAR results.
    """

    # set filenames
    incar_simmate_name = tmp_path / "INCAR"
    incar_pymatgen_name = tmp_path / "INCAR_pmg"

    # Because we won't have POTCARs accessible, we need to cover this function
    # call -- specifically have it pretend to make a file
    potcar_filename = tmp_path / "POTCAR"
    mocker.patch.object(
        Potcar,
        "to_file_from_type",
        return_value=make_dummy_files(potcar_filename),
    )

    # write both inputs
    simmate_workflow.setup(structure=structure, directory=tmp_path)
    pymatgen_set(structure, **pymatgen_kwargs).incar.write_file(incar_pymatgen_name)

    # load incar
    incar_simmate = Incar.from_file(incar_simmate_name)
    incar_pymatgen = Incar.from_file(incar_pymatgen_name)

    # compare
    diff = incar_simmate.compare_incars(incar_pymatgen)["Different"]
    # ignore kpoints input for now
    diff.pop("KSPACING", None)

    # in EDIFF__per_atom, there are sometimes rounding issues.
    # We say there is no difference if they are within 100% error. This may
    # seem like a large error, but it is meant to encompass differences like
    # {'EDIFF': (4e-06, 1e-05)} where pymatgen rounded significantly. There
    # really isn't an issue unless EDIFF is off by a factor of 10 (1000%)
    if "EDIFF" in diff and "EDIFF__per_atom" in simmate_workflow.incar:
        ediff_simmate, ediff_pymatgen = diff["EDIFF"]
        percent_diff = (ediff_pymatgen - ediff_simmate) / ediff_pymatgen
        if abs(percent_diff) < 1:
            diff.pop("EDIFF")

    # ensure incars are the same (there are no differences)
    assert diff == {}


@pytest.mark.pymatgen
@pytest.mark.parametrize(
    "simmate_workflow, pymatgen_set, pymatgen_kwargs",
    [
        (Diffusion__Vasp__NebFromImagesMit, MITNEBSet, {}),
        (Diffusion__Vasp__NebFromImagesMvlCi, MVLCINEBSet, {}),
    ],
)
def test_pymatgen_input_sets_neb(
    structure,
    tmp_path,
    mocker,
    simmate_workflow,
    pymatgen_set,
    pymatgen_kwargs,
):
    """
    This is a copy/paste of the test method above, but this is for sets that
    accept multiple structures as inputs (such as NEB sets)
    """

    # set filenames
    incar_simmate_name = tmp_path / "INCAR"
    incar_pymatgen_name = tmp_path / "INCAR_pmg"

    # Because we won't have POTCARs accessible, we need to cover this function
    # call -- specifically have it pretend to make a file
    potcar_filename = tmp_path / "POTCAR"
    mocker.patch.object(
        Potcar,
        "to_file_from_type",
        return_value=make_dummy_files(potcar_filename),
    )

    # write both inputs. We just copy the same structure 3 times because we
    # don't really care what the POSCARs look like -- just the INCARs
    simmate_workflow.setup(
        migration_images=MigrationImages([structure] * 3),
        directory=tmp_path,
    )
    pymatgen_set([structure] * 3, **pymatgen_kwargs).incar.write_file(
        incar_pymatgen_name
    )

    # load incar
    incar_simmate = Incar.from_file(incar_simmate_name)
    incar_pymatgen = Incar.from_file(incar_pymatgen_name)

    # compare
    diff = incar_simmate.compare_incars(incar_pymatgen)["Different"]
    # ignore kpoints input for now
    diff.pop("KSPACING", None)

    # in EDIFF__per_atom, there are sometimes rounding issues.
    # We say there is no difference if they are within 100% error. This may
    # seem like a large error, but it is meant to encompass differences like
    # {'EDIFF': (4e-06, 1e-05)} where pymatgen rounded significantly. There
    # really isn't an issue unless EDIFF is off by a factor of 10 (1000%)
    if "EDIFF" in diff and "EDIFF__per_atom" in simmate_workflow.incar:
        ediff_simmate, ediff_pymatgen = diff["EDIFF"]
        percent_diff = (ediff_pymatgen - ediff_simmate) / ediff_pymatgen
        if abs(percent_diff) < 1:
            diff.pop("EDIFF")

    # ensure incars are the same (there are no differences)
    assert diff == {}



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.workflows.electronic_structure.base_full import (
    ElectronicStructureWorkflow,
)
from simmate.calculators.vasp.workflows.electronic_structure.matproj_band_structure import (
    ElectronicStructure__Vasp__MatprojBandStructure,
)
from simmate.calculators.vasp.workflows.electronic_structure.matproj_density_of_states import (
    ElectronicStructure__Vasp__MatprojDensityOfStates,
)
from simmate.calculators.vasp.workflows.static_energy.matproj import (
    StaticEnergy__Vasp__Matproj,
)


class ElectronicStructure__Vasp__MatprojFull(ElectronicStructureWorkflow):
    """
    runs DOS and BS at Materials Project settings
    """

    static_energy_workflow = StaticEnergy__Vasp__Matproj
    band_structure_workflow = ElectronicStructure__Vasp__MatprojBandStructure
    density_of_states_workflow = ElectronicStructure__Vasp__MatprojDensityOfStates



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.error_handlers import Brmix
from simmate.calculators.vasp.inputs import Incar
from simmate.conftest import copy_test_files, make_dummy_files


def test_brmix(tmp_path):
    copy_test_files(
        tmp_path,
        test_directory=__file__,
        test_folder="brmix.zip",
    )

    # we reference the files several spots below so we grab its path up front
    incar_filename = tmp_path / "INCAR"
    chgcar_filename = tmp_path / "CHGCAR"
    wavecar_filename = tmp_path / "WAVECAR"
    outcar_filename = tmp_path / "OUTCAR"
    errorcount_filename = tmp_path / "simmate_error_counts.json"

    # init class with default settings
    error_handler = Brmix()

    # Confirm an error IS NOT found
    error_handler.filename_to_check = "vasp.no_error"
    assert error_handler.check(tmp_path) == False

    # Confirm an error IS found
    error_handler.filename_to_check = "vasp.out"
    assert error_handler.check(tmp_path) == True

    # Handle error when there is a valid OUTCAR
    fix = error_handler.correct(tmp_path)
    assert fix == "switched ISTART to 1"
    incar = Incar.from_file(incar_filename)
    assert incar["ISTART"] == 1

    # Remove the error counts and valid OUTCAR and restart
    errorcount_filename.unlink()
    outcar_filename.unlink()

    # Make first attempt at fixing the error
    fix = error_handler.correct(tmp_path)
    assert fix == "switched IMIX to 1"
    incar = Incar.from_file(incar_filename)
    assert incar["IMIX"] == 1

    # Make second attempt at fixing the error
    fix = error_handler.correct(tmp_path)
    assert fix == "removed any IMIX tag and switched KGAMMA to False"
    incar = Incar.from_file(incar_filename)
    assert incar.get("IMIX", None) is None
    assert incar["KGAMMA"] == False

    # Make third attempt at fixing the error
    fix = error_handler.correct(tmp_path)
    assert fix == "removed any IMIX tag and switched KGAMMA to True"
    incar = Incar.from_file(incar_filename)
    assert incar.get("IMIX", None) is None
    assert incar["KGAMMA"] == True

    # Make final attempt at fixing the error
    make_dummy_files(chgcar_filename, wavecar_filename)
    fix = error_handler.correct(tmp_path)
    assert fix == "switched ISYM to 0 and KGAMMA to True and deleted CHGCAR and WAVECAR"
    incar = Incar.from_file(incar_filename)
    assert incar["ISYM"] == 0
    assert incar["KGAMMA"] == True
    assert not chgcar_filename.exists()
    assert not wavecar_filename.exists()

    # Confirm an error IS NOT found when NELECT is in the INCAR
    with incar_filename.open("w") as file:
        file.write("NELECT = 123")
    assert error_handler.check(tmp_path) == False



# -*- coding: utf-8 -*-

import logging

import numpy
import scipy
from django.utils import timezone
from rich.progress import track

from simmate.toolkit import Structure
from simmate.toolkit.validators import Validator


class FingerprintValidator(Validator):

    comparison_mode: str = "linalg_norm"
    """
    How fingerprints distances should be determined. Options are:
        - linalg_norm
        - cos
        - custom

    For 'custom', you must write a custom 'get_fingerprint_distance(fp1, fp2)'
    static method that will be used.
    """

    def __init__(
        self,
        distance_tolerance: float = 0.001,
        structure_pool: list[Structure] = [],  # OR a queryset from a Structure table
        add_unique_to_pool: bool = True,
        **kwargs,
    ):

        # OPTIMIZE: my choice of tolerance is based on my evolutionary search.
        # However, "clustering" after evolutionary search suggests that something
        # like 0.5 might be a better choice for finding a truly unique structure
        self.distance_tolerance = distance_tolerance

        self.add_unique_to_pool = add_unique_to_pool

        # setup featurizer with the given composition
        self.featurizer = self.get_featurizer(**kwargs)

        # check if we were given a list of pymatgen structures.
        if isinstance(structure_pool, list):
            # set this variable as none to help with some warnings methods such
            # as the update_fingerprint_database
            self.structure_pool_queryset = "local_only"
            # If so, we generate the fingerprint for each of the initial input structures
            # We convert this to a numpy array for speed improvement at later stages
            # OPTIMIZE: this can be slow and I should support parallel featurization
            self.fingerprint_database = numpy.array(
                [
                    self.featurizer.featurize(structure)
                    for structure in track(structure_pool)
                ]
            )

        # otherwise we have a queryset that should be used to populate the
        # fingerprint database
        else:
            # we store the queryset as an attribute because we may want to
            # update the structure pool later on.
            self.structure_pool_queryset = structure_pool
            self.fingerprint_database = numpy.array([])
            # we also keep a log of the last update so we only grab new structures
            # each time we update the database. To start, we set this as the
            # eariest possible date, which tells our update_fingerprint_database
            # method to include ALL structures
            self.last_update = timezone.make_aware(
                timezone.datetime.min, timezone.get_default_timezone()
            )
            self.update_fingerprint_database()

    @staticmethod
    def get_featurizer(self, **kwargs):
        raise NotImplementedError(
            "make sure you add a custom 'get_featurizer' method. This should "
            "return a featurizer object to be used for generating fingerprints."
        )

    def get_fingerprint_distance(self, fingerprint1, fingerprint2) -> float:
        raise NotImplementedError(
            "If set the class attribute 'comparison_mode' to 'custom', make "
            "sure you add a custom 'get_fingerprint_distance' method. This should "
            "return a distance (float value) ."
        )

    @staticmethod
    def format_fingerprint(fingerprint):
        return fingerprint  # does nothing by default

    def check_structure(self, structure: Structure):

        # make the fingerprint for this structure and make into a numpy array for speed
        fingerprint1 = numpy.array(self.featurizer.featurize(structure))

        # apply any extra formatting
        fingerprint1 = self.format_fingerprint(fingerprint1)

        # We now want to get the distance of this fingerprint relative to all others.
        # If the distance is within the specified tolerance, then the structures
        # are too similar - and we return  for a failure.
        for fingerprint2 in self.fingerprint_database:

            # check fingerprint based on the mode set
            if self.comparison_mode == "linalg_norm":
                distance = numpy.linalg.norm(fingerprint1 - fingerprint2)
            elif self.comparison_mode == "cos":
                distance = scipy.spatial.distance.cosine(fingerprint1, fingerprint2)
            elif self.comparison_mode == "custom":
                distance = self.get_fingerprint_distance(fingerprint1, fingerprint2)
            else:
                raise NotImplementedError("Unknown comparison_mode provided.")

            # and determine if we have a match
            if distance < self.distance_tolerance:
                # we can end the whole function as soon as one structure is
                # deemed too similar
                return False
            else:
                continue
        # If we make it through all structures and no distance is below the
        # tolerance, then we have a new and unique structure!

        # add this new structure to the database if it was requested.
        if self.add_unique_to_pool:
            self._add_fingerprint_to_database(fingerprint1)

        # Return that we were successful
        return True

    def update_fingerprint_database(self):

        if self.structure_pool_queryset == "local_only":
            raise Exception(
                "This method should only be used when your structure pool"
                " is based on a Simmate database table!"
            )

        # what if a structure was added to the table WHILE I was running the
        # last query? This sets up a race condition with new structures and my
        # last_update timestamp. To avoid this, I immediately update our
        # timestamp, rather than doing it after the query
        last_update_safe = self.last_update
        self.last_update = timezone.now()

        # now grab all the new structures!
        new_structures = (
            self.structure_pool_queryset.filter(created_at__gte=last_update_safe)
            .only("structure")
            .to_toolkit()
        )

        # If there aren't any new structures, just exit without printing the
        # message and progress bar
        if not new_structures:
            return

        logging.info(
            f"Found {len(new_structures)} new structures for the fingerprint database."
        )

        # calculate each fingerprint and add it to the database
        for structure in track(new_structures):
            fingerprint = numpy.array(self.featurizer.featurize(structure))
            self._add_fingerprint_to_database(fingerprint)

    def _add_fingerprint_to_database(self, fingerprint):
        if self.fingerprint_database.size == 0:
            self.fingerprint_database = numpy.array([fingerprint])
        else:
            self.fingerprint_database = numpy.append(
                self.fingerprint_database, [fingerprint], axis=0
            )



# The order that we import these different modules is important to prevent
# circular imports errors, so we prevent isort from changing this file.
# isort: skip_file

from .base import FingerprintValidator

from .rdf import RdfFingerprint
from .prdf import PartialRdfFingerprint
from .crystalnn import CrystalNNFingerprint
from .pcrystalnn import PartialCrystalNNFingerprint



# -*- coding: utf-8 -*-

import logging
import time
from pathlib import Path

from simmate.database.workflow_results import FixedCompositionSearch
from simmate.toolkit import Composition
from simmate.workflow_engine import Workflow


class StructurePrediction__Toolkit__FixedComposition(Workflow):
    """
    Runs an evolutionary search algorithm on a fixed composition with a fixed
    number of sites. (e.g. Ca2N or Na4Cl4)
    """

    database_table = FixedCompositionSearch

    @classmethod
    def run_config(
        cls,
        composition: str | Composition,
        subworkflow_name: str | Workflow = "relaxation.vasp.staged",
        subworkflow_kwargs: dict = {},
        fitness_field: str = "energy_per_atom",
        max_structures: int = None,
        min_structures_exact: int = None,
        best_survival_cutoff: int = None,
        convergence_cutoff: float = 0.001,
        nfirst_generation: int = 15,
        nsteadystate: int = 40,
        singleshot_sources: list[str] = [
            "third_parties",
            "prototypes",
            # "third_party_substituition",
        ],
        steadystate_sources: dict = {
            "RandomSymStructure": 0.60,
            "from_ase.Heredity": 0.25,
            # "from_ase.SoftMutation": 0.10,
            "from_ase.MirrorMutation": 0.05,
            # "from_ase.LatticeStrain": 0.05,
            "from_ase.RotationalMutation": 0.05,
            "from_ase.AtomicPermutation": 0.05,
            # "from_ase.CoordinatePerturbation": 0.05,
            # "ExtremeSymmetry": 0.05,
        },
        selector_name: str = "TruncatedSelection",
        selector_kwargs: dict = {},
        validator_name: str = "PartialRdfFingerprint",
        validator_kwargs: dict = {
            "distance_tolerance": 0.001,
            "cutoff": 10.0,
            "bin_size": 0.1,
        },
        sleep_step: int = 60,
        directory: Path = None,
        write_summary_files: bool = True,
        run_id: str = None,
        **kwargs,
    ):

        #######################################################################
        # TODO: Check if there is an existing search and grab it if so.
        # We can also update, add logs, compare... I need to decide what to do
        # here. The stop condition could also be checked to see if a search
        # even needs to be started.
        # if SearchDatatable.objects.filter(composition=composition.formula).exists():
        #     self.search_datatable = SearchDatatable.objects.filter(
        #         composition=composition.formula,
        #     ).get()
        # Grab the list of singleshot sources that have been ran before
        # and based off of that list, remove repeated sources
        # past_singleshot_sources = (
        #     self.search_datatable.sources.filter(is_singleshot=True)
        #     .values_list("name", flat=True)
        #     .all()
        # )
        # singleshot_sources = [
        #     source
        #     for source in singleshot_sources
        #     if source not in past_singleshot_sources
        # ]
        # BUG: what if I want to rerun any of these even though its been ran before?
        #######################################################################

        logging.info(f"Setting up evolutionary search for {composition}")

        # grab the calculation table linked to this workflow run
        search_datatable = cls.database_table.objects.get(run_id=run_id)

        # Convergence criteria and stop conditions can be set based on the
        # number of atoms in the composition. Here we try to set reasonable criteria
        # for these if a user-input was not given. Note we are using max(..., N)
        # to set an absolute minimum for these.
        n = composition.num_atoms
        min_structures_exact = min_structures_exact or max([int(30 * n), 100])
        max_structures = max_structures or max([int(n * 250 + n**2.75), 1500])
        best_survival_cutoff = best_survival_cutoff or max([int(30 * n + n**2), 100])
        search_datatable.update_from_fields(
            min_structures_exact=min_structures_exact,
            max_structures=max_structures,
            best_survival_cutoff=best_survival_cutoff,
        )

        # sometimes the conditions are already met by a previous search so we
        # check for this up front.
        if search_datatable.check_stop_condition():
            logging.info("Looks like this search was already ran by someone else!")
            return

        logging.info("Finished setup")
        logging.info(
            f"Assigned this to FixedCompositionSearch id={search_datatable.id}."
        )

        # See if the singleshot sources have been ran yet. For restarted calculations
        # this will likely not be needed (unless a new source was added). But for
        # new searches/compositions, this will submit all individuals from the
        # single shot sources before we even start the steady-state runs
        search_datatable._check_singleshot_sources(directory)

        # this loop will go until I hit 'break' below
        while True:

            # Write the output summary if there is at least one structure completed
            if write_summary_files:
                if search_datatable.individuals_completed.count() >= 1:
                    search_datatable.write_summary(directory)
                else:
                    search_datatable.write_individuals_incomplete(directory)

            # TODO: maybe write summary files to csv...? This may be a mute
            # point because I expect we can follow along in the web UI in the future
            # To that end, I can add a "URL" property

            # Check the stop condition
            # If it is True, we can stop the calc.
            if search_datatable.check_stop_condition():
                # TODO: should I sleep / wait for other calculations and
                # check again? Then write summary one last time..?
                break  # break out of the while loop
            # Otherwise, keep going!

            # TODO: Go through triggered actions that would update the database
            # table -- e.g. the workflow to run, the validators, etc.
            # self._check_triggered_actions()

            # Go through the running workflows and see if we need to submit
            # new ones to meet our steadystate target(s)
            search_datatable._check_steadystate_workflows()

            # To save our database load, sleep until we run checks again.
            logging.info(
                f"Sleeping for {sleep_step} seconds before running checks again."
            )
            time.sleep(sleep_step)

        logging.info("Stopping the search (running calcs will be left to finish).")



# -*- coding: utf-8 -*-

import pytest

from simmate.calculators.vasp.inputs.potcar_mappings import PBE_ELEMENT_MAPPINGS
from simmate.calculators.vasp.workflows.base import VaspWorkflow
from simmate.conftest import SimmateMockHelper, copy_test_files


class Testing__Vasp__Dummy(VaspWorkflow):
    """
    A minimal example VaspWorkflow that is just for testing
    """

    use_database = False

    functional = "PBE"
    potcar_mappings = PBE_ELEMENT_MAPPINGS
    standardize_structure = "primitive-LLL"
    symmetry_tolerance = 0.1
    angle_tolerance = 10.0

    incar = dict(
        PREC="Low",
        EDIFF__per_atom=2e-3,  # to ensure structure-specific kwargs
        KSPACING=0.75,
    )


# For shorthand reference below
DummyWorkflow = Testing__Vasp__Dummy


def test_base_setup(structure, tmp_path, mocker):

    Potcar = SimmateMockHelper.get_mocked_potcar(mocker, tmp_path)

    # estabilish filenames that we make and commonly reference
    incar_filename = tmp_path / "INCAR"
    poscar_filename = tmp_path / "POSCAR"
    potcar_filename = tmp_path / "POTCAR"

    # try to make input files in the tmp_path
    DummyWorkflow.setup(directory=tmp_path, structure=structure)
    assert incar_filename.exists()
    assert poscar_filename.exists()
    assert potcar_filename.exists()
    Potcar.to_file_from_type.assert_called_with(
        structure.composition.elements,
        "PBE",
        potcar_filename,
        PBE_ELEMENT_MAPPINGS,
    )


def test_base_vasp_run(structure, tmp_path, mocker):
    copy_test_files(
        tmp_path,
        test_directory=__file__,
        test_folder="base.zip",
    )

    SimmateMockHelper.mock_vasp(mocker)

    # estabilish filenames that we make and commonly reference
    vasprun_filename = tmp_path / "vasprun.xml"

    # run the full workflow
    DummyWorkflow.run(structure=structure, directory=tmp_path)

    # run the workup again with a malformed xml
    with vasprun_filename.open("r") as file:
        contents = file.readlines()
    with vasprun_filename.open("w") as file:
        file.writelines(contents[50])
    with pytest.raises(Exception):
        DummyWorkflow.run(tmp_path)



# -*- coding: utf-8 -*-

from pathlib import Path

from simmate.calculators.vasp.workflows.diffusion.utilities import (
    get_migration_images_from_endpoints,
)
from simmate.toolkit import Structure
from simmate.workflow_engine import Workflow


class NebFromEndpointWorkflow(Workflow):
    """
    Runs a full diffusion analysis from a start and end supercell using NEB.

    The input start/end structures will be relaxed using the relaxation/neb-endpoint
    workflow and then interpolated to generate 7 images. These 7 images are
    then relaxed using NEB within the diffusion/from-images workflow.

    This is therefore a "Nested Workflow" made of the following smaller workflows:

        - relaxation/neb-endpoint (for both start and end supercells)
        - a mini task that interpolated relaxed endpoints and makes images
        - diffusion/from-images

    If you are running this workflow via the command-line, you can run this
    with...

    ``` bash
    simmate workflows run diffusion/all-paths -c "cmd1; cmd2" --supercell_start example1.cif --supercell_end example2.cif
    ```

    Note, the `-c` here is very important! Here we are passing three commands
    separated by semicolons. Each command is passed to a specific workflow call:

        - cmd1 --> used for endpoint supercell relaxations
        - cmd2 --> used for NEB

    Thus, you can scale your resources for each step. Here's a full -c option:

    ```
    -c "mpirun -n 12 vasp_std > vasp.out; mpirun -n 70 vasp_std > vasp.out"
    ```
    """

    description_doc_short = "runs NEB using two endpoint structures as input"

    # Oddly enough, the from_images_workflow and this workflow share a table
    # entry, so nothing needs to be done for working up results. See
    # how we pass run_id=run_id below.
    update_database_from_results = False

    endpoint_relaxation_workflow: Workflow = None
    from_images_workflow: Workflow = None

    @classmethod
    def run_config(
        cls,
        supercell_start: Structure,
        supercell_end: Structure,
        directory: Path = None,
        source: dict = None,
        command: str = None,
        nimages: int = 5,
        diffusion_analysis_id: int = None,
        is_restart: bool = False,
        run_id: str = None,
        **kwargs,
    ):
        # command list expects three subcommands:
        #   command_bulk, command_supercell, and command_neb
        # subcommands = parse_multi_command(
        #     command,
        #     commands_out=["command_supercell", "command_neb"],
        # )

        # Relax the starting supercell structure
        endpoint_start_state = cls.endpoint_relaxation_workflow.run(
            structure=supercell_start,
            command=command,  # subcommands["command_supercell"]
            directory=directory / f"{cls.endpoint_relaxation_workflow.name_full}.start",
            is_restart=is_restart,
        )

        # Relax the ending supercell structure
        endpoint_end_state = cls.endpoint_relaxation_workflow.run(
            structure=supercell_end,
            command=command,  # subcommands["command_supercell"]
            directory=directory / f"{cls.endpoint_relaxation_workflow.name_full}.end",
            is_restart=is_restart,
        )

        # wait for the endpoint relaxations to finish
        endpoint_start_result = endpoint_start_state.result()
        endpoint_end_result = endpoint_end_state.result()

        # interpolate / empirically relax the endpoint structures
        images = get_migration_images_from_endpoints(
            supercell_start=endpoint_start_result,
            supercell_end=endpoint_end_result,
            nimages=nimages,
        ).result()

        # Run NEB on this set of images
        neb_state = cls.from_images_workflow.run(
            migration_images=images,
            command=command,  # subcommands["command_neb"]
            directory=directory,
            diffusion_analysis_id=diffusion_analysis_id,
            is_restart=is_restart,
            # Run id is very important here as it tells the underlying
            # workflow that it doesn't need to create a new database object
            # during registration -- as it will use the one that was registered
            # when this workflow started. This is also why we have a dummy
            # `update_database_from_results` method below
            run_id=run_id,
        )

        return neb_state.result()



# -*- coding: utf-8 -*-

# The order that we import these different modules is important to prevent
# circular imports errors, so we prevent isort from changing this file.
# isort: skip_file

from .base import DatabaseTable, table_column, SearchResults

from .symmetry import Spacegroup
from .structure import Structure
from .calculation import Calculation
from .forces import Forces
from .thermodynamics import Thermodynamics

from .static_energy import StaticEnergy
from .relaxation import Relaxation, IonicStep
from .dynamics import Dynamics, DynamicsIonicStep
from .calculation_nested import NestedCalculation
from .band_structure import BandStructure, BandStructureCalc
from .density_of_states import DensityofStates, DensityofStatesCalc
from .population_analysis import PopulationAnalysis

from .nudged_elastic_band import DiffusionAnalysis, MigrationHop, MigrationImage



# -*- coding: utf-8 -*-

"""
This defines commands for managing your Simmate workflow engine. All commands are 
accessible through the "simmate workflow-engine" command.
"""

import typer

workflow_engine_app = typer.Typer(rich_markup_mode="markdown")


@workflow_engine_app.callback(no_args_is_help=True)
def workflow_engine():
    """
    A group of commands for starting up computational resources (Workers,
    Agents, and Clusters)
    """
    pass


@workflow_engine_app.command()
def start_worker(
    nitems_max: int = None,
    timeout: float = None,
    close_on_empty_queue: bool = False,
    waittime_on_empty_queue: float = 1,
    tag: list[str] = ["simmate"],
):
    """
    Starts a Simmate Worker which will query the database for jobs to run

    By default this worker will run endlessly and not close.

    - `nitems_max`: the number of task run to submit before shutdown
    - `timeout`: the time (in seconds) after which this worker will stop running
    jobs and shutdown

    - `close_on_empty_queue`: whether the worker should shut down when the
    queue is empty

    - `waittime_on_empty_queue`: if the queue is empty, the time (in seconds)
    the worker should wait before checking the queue again

    - `tags`: tags to filter tasks by for submission. defaults to just 'simmate'

    """

    from simmate.workflow_engine import Worker

    worker = Worker(
        nitems_max,
        timeout,
        close_on_empty_queue,
        waittime_on_empty_queue,
        tag,  # this is actually "tags" --> a list of strings
    )
    worker.start()


@workflow_engine_app.command()
def start_singleflow_worker():
    """
    This starts a Simmate Worker that only runs one job and then shuts down. Also,
    if no job is available in the queue, it will shut down.

    Note: this can be acheived using the start-worker command too, but because
    this is such a common use-case, we include this command for convienence.
    It is the same as...

    `simmate workflow-engine start-worker --nitems-max 1 --close-on-empty-queue`
    """

    from simmate.workflow_engine import Worker

    Worker.run_singleflow_worker()


@workflow_engine_app.command()
def start_cluster(
    nworkers: int,
    worker_command: str = "simmate workflow-engine start-worker",
):
    """
    This starts many Simmate Workers that each run in a local subprocess

    - `nworkers`: the number of workers to start

    - `worker_command`: the command to start each worker with. See start-worker
    for options

    """

    from simmate.workflow_engine.execution.utilities import start_cluster

    start_cluster(
        nworkers=nworkers,
        worker_command=worker_command,
    )


@workflow_engine_app.command()
def show_error_summary():
    """
    Prints the shorthand error for all failed jobs
    """
    from simmate.workflow_engine.execution import SimmateExecutor

    SimmateExecutor.show_error_summary()



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.workflows.static_energy.matproj import (
    StaticEnergy__Vasp__Matproj,
)


class VaspDensityOfStates(StaticEnergy__Vasp__Matproj):
    """
    A base class for density of states (DOS) calculations. This is not meant
    to be used directly but instead should be inherited from.

    This is also a non self-consistent field (non-SCF) calculation and thus uses
    the a fixed charge density from a previous static energy calculation.
    """

    required_files = StaticEnergy__Vasp__Matproj.required_files + ["CHGCAR"]



# -*- coding: utf-8 -*-

from pathlib import Path

from simmate.calculators.bader.workflows import (
    PopulationAnalysis__Bader__Bader,
    PopulationAnalysis__Bader__CombineChgcars,
)
from simmate.calculators.vasp.workflows.static_energy.matproj import (
    StaticEnergy__Vasp__Matproj,
)
from simmate.toolkit import Structure
from simmate.workflow_engine import Workflow


class PopulationAnalysis__VaspBader__BaderMatproj(Workflow):
    """
    Runs a static energy calculation using an extra-fine FFT grid and then
    carries out Bader analysis on the resulting charge density.
    """

    @classmethod
    def run_config(
        cls,
        structure: Structure,
        command: str = None,
        source: dict = None,
        directory: Path = None,
        **kwargs,
    ):

        prebader_result = StaticEnergy__Vasp__PrebaderMatproj.run(
            structure=structure,
            command=command,
            source=source,
            directory=directory,
        ).result()

        # Setup chargecars for the bader analysis and wait until complete
        PopulationAnalysis__Bader__CombineChgcars.run(
            directory=prebader_result.directory,
        ).result()

        # Bader only adds files and doesn't overwrite any, so I just run it
        # in the original directory. I may switch to copying over to a new
        # directory in the future though.
        PopulationAnalysis__Bader__Bader.run(
            directory=prebader_result.directory,
        ).result()


class StaticEnergy__Vasp__PrebaderMatproj(StaticEnergy__Vasp__Matproj):
    """
    Runs a static energy calculation with a high-density FFT grid under settings
    from the Materials Project. Results can be used for Bader analysis.

    We do NOT recommend running this calculation on its own. Instead, you should
    use the full workflow, which runs this calculation AND the following bader
    analysis for you. This S3Task is only the first step of that workflow.

    See `bader.workflows.materials_project`.
    """

    # The key thing for bader analysis is that we need a very fine FFT mesh. Other
    # than that, it's the same as a static energy calculation.
    incar = StaticEnergy__Vasp__Matproj.incar.copy()
    incar.update(
        NGXF__density_a=20,
        NGYF__density_b=20,
        NGZF__density_c=20,
        LAECHG=True,  # write core charge density to AECCAR0 and valence to AECCAR2
    )



# -*- coding: utf-8 -*-

"""
WARNING: This module is still at the planning stage and the code below serve as
placeholder notes.

In the future, it will be a table that helps link together results from
multi-stage workflows.
"""

from simmate.database.base_data_types import Calculation  # , table_column


class NestedCalculation(Calculation):
    """
    A nested calculation is extremely similar to a calculation table, except here
    the linked workflow is a NestedWorkflow -- meaning it is made of multiple
    smaller workflows. For example, we may want to run a workflow that runs a
    series of relaxations. Or maybe a relaxation, then energy, then band-strucuture
    calculation. This table is for keeping track of workflows ran in series like
    this.

    Typically, you'll use the `create_subclass_from_calcs` method to create a
    subclass of this table, which handles making all the relationships for you.
    """

    class Meta:
        abstract = True
        app_label = "workflows"

    # TODO:
    # child_database_tables = [...]

    # TODO:
    # should this be a list of all modifications? It could maybe be used to
    # carry fixes (such as smearing) accross different calcs.
    # Or maybe a method that just lists the corrections of each subcalc?
    corrections = None
    # For now I delete this column. This line removes the field.

    # def update_calculation(self):
    #     """
    #     This is an experimental method that iterates through child workflow tables
    #     and links the results to the NestedCalculation
    #     """
    #     # BUG: This assumes we ran all calculations within the same directory,
    #     # which isn't true in all cases.
    #     for child_calc_table in self.child_database_tables:
    #         if child_calc_table.objects.filter(directory=self.directory).exists():
    #             child_calc = child_calc_table.objects.get(directory=self.directory)
    #             setattr(self, child_calc._meta.model_name, child_calc)
    #     self.save()



# -*- coding: utf-8 -*-

from pathlib import Path

from simmate.visualization.plotting import Figure


class PlotlyFigure(Figure):
    @classmethod
    def view_plot(cls, parent_class, **kwargs):
        figure = cls.get_plot(parent_class, **kwargs)
        cls.apply_theme(figure)
        figure.show(renderer="browser")

    @classmethod
    def write_plot(cls, parent_class, directory: Path = None, **kwargs):
        if not directory:
            directory = Path.cwd()
        figure = cls.get_plot(parent_class, **kwargs)
        cls.apply_theme(figure)
        figure.write_html(
            directory / f"{cls.name}.html",
            include_plotlyjs="cdn",
        )

    @classmethod
    def get_html_div(cls, parent_class, **kwargs):
        # Make the convergence figure and convert it to an html div
        figure = cls.get_plot(parent_class, **kwargs)
        cls.apply_theme(figure)
        html_div = figure.to_html(
            full_html=False,
            include_plotlyjs=False,
        )
        return html_div

    @staticmethod
    def apply_theme(plot):
        plot.update_layout(template="ggplot2")



# -*- coding: utf-8 -*-

"""
This file only exists because DigitalOcean and Heroku require a setup.py
to detect that python is being used. Once they update to pyproject.toml,
this file will be removed.
"""

from setuptools import setup

setup()



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.workflows.diffusion.neb_all_paths_base import (
    NebAllPathsWorkflow,
)
from simmate.calculators.vasp.workflows.diffusion.neb_single_path_mit import (
    Diffusion__Vasp__NebSinglePathMit,
)
from simmate.calculators.vasp.workflows.relaxation.mit import Relaxation__Vasp__Mit
from simmate.calculators.vasp.workflows.static_energy.mit import StaticEnergy__Vasp__Mit


class Diffusion__Vasp__NebAllPathsMit(NebAllPathsWorkflow):
    """
    Runs a full diffusion analysis on a bulk crystal structure using NEB with
    MIT settings.
    """

    bulk_relaxation_workflow = Relaxation__Vasp__Mit
    bulk_static_energy_workflow = StaticEnergy__Vasp__Mit
    single_path_workflow = Diffusion__Vasp__NebSinglePathMit



# -*- coding: utf-8 -*-

"""
This module defines the lowest-level classes for database tables and their
search results.

See the `simmate.database.base_data_types` (which is the parent module of 
this one) for example usage.
"""

import inspect
import json
import logging
import shutil
import urllib
import warnings
from functools import cache
from pathlib import Path

import pandas
import yaml
from django.db import models  # , transaction
from django.db import models as table_column
from django.utils.module_loading import import_string
from django.utils.timezone import datetime
from django_filters import rest_framework as django_api_filters
from django_pandas.io import read_frame
from rich.progress import track

# The "as table_column" line does NOTHING but rename a module.
# I have this because I want to use "table_column.CharField(...)" instead
# of "models.CharField(...)" in my Models. This let's beginners read my
# higher level classes and instantly understand what each thing represents
# -- without them needing to understand that Django Model == Database Table.
# Experts may find this annoying, so I'm sorry :(


class SearchResults(models.QuerySet):
    """
    This class adds some extra methods to the results returned from a database
    search. For example, if you searched all Structures and wanted to convert
    these to a pandas dataframe or even a list of pymatgen structures, you can
    now do...

    ``` python
    # for a list of database objects (django models)
    search_results = Structures.objects.all()

    # for a pandas dataframe (which is like an Excel table)
    dataframe = search_results.to_dataframe()

    # for a list of pymatgen structure objects
    structures = search_results.to_toolkit()
    ```

    All other functionality is inherited from
    [Django QuerySets](https://docs.djangoproject.com/en/4.0/ref/models/querysets/).
    """

    def to_dataframe(
        self,
        fieldnames: list[str] = (),
        verbose: bool = True,
        index: str = None,
        coerce_float: str = False,
        datetime_index: str = False,
    ) -> pandas.DataFrame:
        """
        Returns a Pandas DataFrame of the search results

        This method is coppied from django_pandas'
        [manager.py](https://github.com/chrisdev/django-pandas/blob/master/django_pandas/managers.py)

        Paramaters
        -----------
        - `fieldnames`:
            The model field names(columns) to utilise in creating the DataFrame.
            You can span a relationships in the usual Django ORM way by using
            the foreign key field name separated by double underscores and refer
            to a field in a related model.
        - `index`:
            specify the field to use  for the index. If the index field is not
            in fieldnames it will be appended. This is mandatory for timeseries.
        - `verbose`:
            If  this is `True` then populate the DataFrame with the human
            readable versions for foreign key fields else use the actual values
            set in the model
        - `coerce_float`:
            Attempt to convert values to non-string, non-numeric objects (like
            decimal.Decimal) to floating point.
        - `datetime_index`: bool
            specify whether index should be converted to a DateTimeIndex.
        """

        # BUG: read_frame runs a NEW query, so it may be a different length from
        # the original queryset.
        # See https://github.com/chrisdev/django-pandas/issues/138 for issue
        return read_frame(
            self,
            fieldnames=fieldnames,
            verbose=verbose,
            index_col=index,
            coerce_float=coerce_float,
            datetime_index=datetime_index,
        )

    def to_toolkit(
        self,
    ) -> list:  # type of object varies (e.g. Structure, BandStructure, etc.)
        """
        Converts your SearchResults to a list of pymatgen objects
        """

        # This method will only be for structures and other classes that
        # support this method. So we make sure the model has supports it first.
        if not hasattr(self.model, "to_toolkit"):
            raise Exception(
                "This database table does not have a to_toolkit method implemented"
            )

        # now we can iterate through the queryset and return the converted
        # pymatgen objects as a list
        return [obj.to_toolkit() for obj in self]

    def to_archive(self, filename: Path | str = None):
        """
        Writes a compressed zip file using the table's `archive_fieldset`
        attribute. Underneath, the file is written in a csv format.

        This is useful for small making archive files and reloading fixtures
        to a separate database.

        This method is attached to the table manager for scenarios to allow
        queryset filtering before dumping data.

        To load this database dump into a new database, use the class's
        `from_archive` method.

        #### Parameters

        - `filename`:
            The filename to write the zip file to. By defualt, None will make
            a filename named MyExampleTableName-2022-01-25.zip, where the date
            will be the current day (for versioning).
        """

        # Generate the file name if one wasn't given.
        if not filename:
            # This is automatically the name of the table plus the date, where
            # the date is for versioning. For example...
            #   MyExampleTable-2022-01-25
            today = datetime.today()
            filename_base = "-".join(
                [
                    self.model.table_name,
                    str(today.year),
                    str(today.month).zfill(2),
                    str(today.day).zfill(2),
                ]
            )
            filename = filename_base + ".zip"

        # convert to path obj
        filename = Path(filename)

        # grab the list of fields that we want to store
        fieldset = self.model.archive_fieldset

        # We want to load the entire table, but only grab the fields that
        # we will be storing in the archive.
        base_objs = self.only(*fieldset)

        # now convert these objects to a pandas dataframe, again using just
        # the archive columns
        df = base_objs.to_dataframe(fieldnames=fieldset)

        # Write the data to a csv file
        # OPTIMIZE: is there a better format that pandas can write to?
        csv_filename = filename.with_suffix(".csv")
        df.to_csv(csv_filename, index=False)

        # now convert the dump file to a compressed zip. In the complex, os
        # functions below we are just grabbing the filename without the
        # zip ending. We are also grabbing the directory that the csv is
        # located in
        shutil.make_archive(
            base_name=filename.with_suffix(""),  # removes .zip ending
            format="zip",
            root_dir=csv_filename.absolute().parent,
            base_dir=csv_filename.name,
        )

        # we can now delete the csv file
        csv_filename.unlink()

    def filter_by_tags(self, tags: list[str]):
        """
        A utility filter() method that
        """

        if tags:
            new_query = self
            for tag in tags:
                new_query = new_query.filter(tags__icontains=tag)
        else:
            new_query = self.filter(tags=[])
        return new_query


# Copied this line from...
# https://github.com/chrisdev/django-pandas/blob/master/django_pandas/managers.py
# It simply converts this queryset class to a manager.
DatabaseTableManager = models.Manager.from_queryset(SearchResults)


class DatabaseTable(models.Model):
    """
    The base class for defining a table in the Simmate database. All tables and
    mixins inherit from this class.

    Usage is identical to
    [models in Django](https://docs.djangoproject.com/en/4.0/#the-model-layer)
    where this class only adds extra methods for convenience.
    """

    class Meta:
        abstract = True

    created_at = table_column.DateTimeField(
        auto_now_add=True,
        blank=True,
        null=True,
    )
    """
    Timestamp of when this row was first added to the database table
    """

    updated_at = table_column.DateTimeField(
        auto_now=True,
        blank=True,
        null=True,
    )
    """
    Timestamp of when this row was was lasted changed / updated
    """

    source = table_column.JSONField(blank=True, null=True)
    """
    > Note, this field is highly experimental at the moment and subject to
    change.
    
    This column indicates where the data came from, and it could be a number 
    of things including...
     - a third party id
     - a structure from a different Simmate datbase table
     - a transformation of another structure
     - a creation method
     - a custom submission by the user
    
    By default, this is a JSON field to account for all scenarios, but some
    tables (such as those in `simmate.database.third_parties`) this is value
    should be the same for ALL entries in the table and therefore the column is
    overwritten as an attribute.
    
    For other tables where this is not a constant, here are some examples of
    values used in this column:
    
    ``` python
    # from a thirdparty database or separate table
    source = {
        "table": "MatprojStructure",
        "id": "mp-123",
    }
    
    # from a random structure creator
    source = {"method": "PyXtalStructure"}
    
    # from a templated structure creator (e.g. substituition or prototypes)
    source = {
        "method": "PrototypeStructure",
        "table": "AFLOWPrototypes",
        "id": 123,
    }
    
    # from a transformation
    source = {
        "method": "MirrorMutation",
        "table": "MatprojStructure",
        "id": "mp-123",
    }
    
    # from a multi-structure transformation
    source = {
        "method": "HereditaryMutation",
        "table": "MatprojStructure",
        "ids": ["mp-123", "mp-321"],
    }
    ```
    """
    # TODO: Explore polymorphic relations instead of a JSON dictionary.
    # Making relationships to different tables makes things difficult to use, so
    # these columns are just standalone.
    #
    # This is will be very important for "source" and "parent_nested_calculations"
    # fields because I have no way to efficiently convert these fields to the objects
    # that they refer to. There's also no good way to access a structure's "children"
    # (i.e. structure where they are the source).
    #
    # I should investigate generic relations in django though:
    # https://docs.djangoproject.com/en/3.2/ref/contrib/contenttypes/#generic-relations
    #
    # Another option is using django-polymorphic.
    # https://django-polymorphic.readthedocs.io/en/latest/
    # This thread is really helpful on the subject:
    # https://stackoverflow.com/questions/30343212/
    #
    # TODO: Consider adding some methods to track the history of a structure.
    #  This would be useful for things like evolutionary algorithms.
    # get_source_parent:
    #   this would iterate through sources until we find one in the same table
    #   as this one. Parent sources are often the most recent transformation
    #   or mutation applied to a structure, such as a MirrorMutation.
    # get_source_seed:
    #   this would iterate through sources until we hit a dead-end. So the seed
    #   source would be something like a third-party database, a method that
    #   randomly create structures, or a prototype.
    # Both of these get more complex when we consider transformation that have
    # multiple parents (and therefore multiple seeds too). An example of this
    # is the HereditaryMutation.

    source_doi: str = None
    """
    Source paper that must be referenced if this data is used. If this is None,
    please refer to the `source` attribute for further details on what to 
    reference.
    """

    remote_archive_link: str = None
    """
    The URL that is used to download the archive and then populate this table.
    Many tables, such as those in `simmate.database.third_parties`, have
    pre-existing data that you can download and load into your local database,
    so if this attribute is set, you can use the `load_remote_archive` method.
    """

    # I override the default manager with the one we define above, which has
    # extra methods useful for our querysets.
    objects = DatabaseTableManager()
    """
    Accesses all of the rows in this datatable and initiates a SearchResults
    object. Using this, you can perform complex filtering and conversions on
    data from this table.
    """

    archive_fields: list[str] = []
    """
    The base information for this database table and only these fields are stored
    when the `to_archive` method is used. Columns excluded from this list can 
    be calculated quickly and will therefore not be stored. Therefore, this list
    can be thought of as the "raw data".
    
    The columns from mix-ins will automatically be added. If you would like to
    remove one of these columns, you can add "--" to the start of the column
    name (e.g. "--energy" will not store the energy).
    
    To see all archive fields, see the `archive_fieldset` property.
    """

    api_filters: dict = {}
    """
    Configuration of fields that can be filtered in the REST API and website 
    interface. This follows the format used by django-filters.
    
    For example...
    ``` python
    filter_fields = {
        column1=["exact"],
        column2=["range"],
        column3=BooleanFilter(...),
    }
    ```
    
    See the `api_filterset` property for the final filter object.
    """

    exclude_from_summary: list[str] = []
    """
    When writing output summaries, these columns will be ignored. This is useful
    if you have a column for storing raw data that isn't friendly to read in
    the yaml format. 'structure' is an example of a field we'd want to
    exclude because its not very readable and is available elsewhere.
    """

    # -------------------------------------------------------------------------
    # Core methods accessing key information and writing summary files
    # -------------------------------------------------------------------------

    @classmethod
    @property
    def table_name(cls) -> str:
        """
        Returns the name of this database table, which will simply match the
        class name.

        Note, we use "table_name" instead of "name" because users may want
        a column titled "name", which would break features. To reduce
        occurance of of this issue, we use "table_name" instead.
        """
        return cls.__name__

    @classmethod
    def get_column_names(cls) -> list[str]:
        """
        Returns a list of all the column names for this table and indicates which
        columns are related to other tables. This is primarily used to help
        view what data is available.
        """
        return [column.name for column in cls._meta.get_fields()]

    @classmethod
    def show_columns(cls):
        """
        Prints a list of all the column names for this table and indicates which
        columns are related to other tables. This is primarily used to help users
        interactively view what data is available.
        """
        # Iterate through and grab the columns. Note we don't use get_column_names
        # here because we are attaching relation data as well.
        column_names = [
            column.name + f" (relation to {column.related_model.table_name})"
            if column.is_relation
            else column.name
            for column in cls._meta.get_fields()
        ]

        # Then use yaml to make the printout pretty (no quotes and separate lines)
        print(yaml.dump(column_names))

    @classmethod
    def get_mixins(cls) -> list:  # -> List[DatabaseTable]
        """
        Grabs the mix-in Tables that were used to make this class. This will
        be mix-ins like Structure, Forces, etc. from the
        `simmate.database.base_data_types` module.
        """
        # this must be imported locally because it depends on all other classes
        # from this module -- and will create circular import issues if outside
        from simmate.database import base_data_types as simmate_mixins

        return [
            parent
            for parent in cls.__bases__
            if hasattr(simmate_mixins, parent.table_name)
            and parent.table_name != "DatabaseTable"
        ]

    @classmethod
    def get_mixin_names(cls) -> list[str]:
        """
        Grabs the mix-in Tables that were used to make this class and returns
        a list of their names.
        """
        return [mixin.table_name for mixin in cls.get_mixins()]

    @classmethod
    def get_extra_columns(cls) -> list[str]:
        """
        Finds all columns that aren't covered by the supported Table mix-ins.

        For example, a table made from...

        ``` python
        from simmate.database.base_data_types import (
            table_column,
            Structure,
            Forces,
        )

        class ExampleTable(Structure, Forces):
            custom_column1 = table_column.FloatField()
            custom_column2 = table_column.FloatField()
        ```

        ... would return ...

        ``` python
        ["custom_column1", "custom_column2"]
        ```
        """

        all_columns = cls.get_column_names()
        columns_w_mixin = [
            column for mixin in cls.get_mixins() for column in mixin.get_column_names()
        ]
        extra_columns = [
            column
            for column in all_columns
            if column not in columns_w_mixin and column != "id"
        ]
        return extra_columns

    def write_output_summary(self, directory: Path):
        """
        This writes a "simmate_summary.yaml" file with key output information.
        """

        fields_to_exclude = self.exclude_from_summary + [
            field for mixin in self.get_mixins() for field in mixin.exclude_from_summary
        ]

        all_data = {}
        for column, value in self.__dict__.items():
            if (
                value != None
                and not column.startswith("_")
                and column not in fields_to_exclude
            ):
                all_data[column] = value

        # also add the table name and entry id and website URL
        all_data["_DATABASE_TABLE_"] = self.table_name
        all_data["_TABLE_ID_"] = self.id

        # EXPERIMNETAL: adding workflow URL
        try:
            from simmate.database.base_data_types import Calculation
            from simmate.workflows.utilities import get_workflow

            if isinstance(self, Calculation):
                workflow = get_workflow(self.workflow_name)
                all_data["_WEBSITE_URL_"] = (
                    "http://127.0.0.1:8000/workflows/"  # I assume local host for now
                    f"{workflow.name_type}/{workflow.name_calculator}/"
                    f"{workflow.name_preset}/{self.id}"
                )
        except:
            pass

        summary_filename = directory / "simmate_summary.yaml"
        with summary_filename.open("w") as file:
            content = yaml.dump(all_data)
            file.write(content)

    def to_dict(self):
        return {
            "database_table": self.table_name,
            "database_id": self.id,
        }

    @staticmethod
    def from_dict(source_dict: dict):

        # This method can be return ANY table, so we need to import all of them
        # here. This is a local import to prevent circular import issues.
        from simmate.website.workflows import models as all_datatables

        # start by loading the datbase table, which is given as a module path
        datatable_str = source_dict["database_table"]

        # Import the datatable class -- how this is done depends on if it
        # is from a simmate supplied class or if the user supplied a full
        # path to the class
        # OPTIMIZE: is there a better way to do this?
        if hasattr(all_datatables, datatable_str):
            datatable = getattr(all_datatables, datatable_str)
        else:
            datatable = import_string(datatable_str)

        # These attributes tells us which structure to grab from our datatable.
        # The user should have only provided one -- if they gave more, we just
        # use whichever one comes first.
        run_id = source_dict.get("run_id")
        database_id = source_dict.get("database_id")
        directory = source_dict.get("directory")

        # we must have either a run_id or database_id
        if not run_id and not database_id and not directory:
            raise Exception(
                "You must have either a run_id, database_id, "
                "or directory provided if you want to load a database entry "
                "from a dictionary of metadata."
            )

        # now query the datable with which whichever was provided. Each of these
        # are unique so all three should return a single calculation.
        if database_id:
            database_object = datatable.objects.get(id=database_id)
        elif run_id:
            database_object = datatable.objects.get(run_id=run_id)
        elif directory:
            database_object = datatable.objects.get(directory=directory)

        return database_object

    # -------------------------------------------------------------------------
    # Methods for loading results from files
    # -------------------------------------------------------------------------

    @classmethod
    def from_directory(cls, directory: Path, as_dict: bool = False):
        """
        Loads data from a directory of files
        """

        # check if we have a VASP directory
        vasprun_filename = directory / "vasprun.xml"
        if vasprun_filename.exists():
            return cls.from_vasp_directory(directory, as_dict=as_dict)

        # TODO: add new elif statements when I begin adding new calculators.

        # If we don't detect any directory, we return an empty dictionary.
        # We don't print a warning or error for now because users may want
        # to populate data entirely in python.
        return {} if as_dict else None

    @classmethod
    def from_vasp_directory(cls, directory: Path, as_dict: bool = False):

        from simmate.calculators.vasp.outputs import Vasprun

        vasprun = Vasprun.from_directory(directory)
        return cls.from_vasp_run(vasprun, as_dict=as_dict)

    # -------------------------------------------------------------------------
    # Methods that handle updating a database entry and its related entries
    # -------------------------------------------------------------------------

    def update_from_fields(self, **fields_to_update):
        # go through each key in the dictionary and attach the value to the
        # attribute of this entry
        for key, value in fields_to_update.items():
            setattr(self, key, value)

        # Now we have all data loaded and attached to the database entry, we
        # can call save() to actually update the database
        self.save()

    def update_from_toolkit(self, **fields_to_update):
        """
        Given fundamental "base info" and toolkit objects, this method will
        populate all relevant columns.

        Note, base info also corresponds to the `archive_fieldset`, which
        represents raw data.

        This method is meant for updating existing database entries with new
        data. If your creating a brand-new database entry, use the
        `from_toolkit` method instead.
        """
        fields_expanded = self.from_toolkit(as_dict=True, **fields_to_update)
        self.update_from_fields(**fields_expanded)

    def update_from_directory(self, directory: Path):
        # This is for simple tables. If there are related table entries that
        # need to be created/updated, then this method should be overwritten.
        data_from_dir = self.from_directory(directory, as_dict=True)
        self.update_from_toolkit(**data_from_dir)

    def update_from_results(self, results: dict, directory: Path):
        """
        Updates a database from the results of a workflow run.

        Typically this method is not called directly, as it is used within
        `Workflow._save_to_database` automatically.
        """

        # First update using the results dictionary
        self.update_from_toolkit(directory=str(directory), **results)

        # Many calculations and datatables will have a "from_directory" method
        # that loads data from files. We use this to grab extra fields and
        # add them to our results.
        self.update_from_directory(directory)

    # -------------------------------------------------------------------------
    # Methods that handle updating a database entry and its related entries
    # -------------------------------------------------------------------------

    @classmethod
    def from_toolkit(cls, as_dict: bool = False, **kwargs):
        """
        Given fundamental "raw data" and toolkit objects, this method
        will populate all relevant columns.

        If the table is made up of multiple mix-ins, this method iterates through
        each of the `_from_toolkit` methods of those mixins and combines the results.
        Therefore, you must view this method for each mix-ins to determine which
        kwargs must be passed.

        #### Parameters

        - `as_dict` :
            Whether to return the populated data as a dictionary or to initialize
            it as a database object. Defaults to False.
        - `**kwargs` :
            All fields required in order to initialize the database entry. For
            example, a Structure table would require a toolkit structure, while
            a Thermodynamics table would require an energy.

        Returns
        -------
        a dictionary if as_dict is True; and a database object if as_dict is False
        """

        # Grab a list of all parent classes
        parents = inspect.getmro(cls)

        # As we go through the parent classes below, we will identify the mixins
        # and populate data using those classes. All of this is fed in to our
        # main class at the end of the function. We keep this running dictionary
        # as we go.
        all_data = kwargs.copy()
        # TODO: How should I best handle passing extra kwargs to the final class
        # initialization? For example, I would want to pass `energy` but I wouldn't
        # want to pass the toolkit structure object. I may update this line in
        # the future to remove python objects. For now, I only remove structure
        # and migration_hop because I know that it is a toolkit object -- not
        # a database column.
        all_data.pop("structure", None)
        all_data.pop("migration_hop", None)
        all_data.pop("migration_images", None)
        all_data.pop("band_structure", None)
        all_data.pop("density_of_states", None)

        for parent in parents:

            # Skip the parent class if it doesn't directly inherit from the
            # DatabaseTable class. We do this because we want the fundamental
            # mixins that come with Simmate (such as Structure, Forces, Thermodynamics).
            # We also skip the classes that don't have a _from_toolkit method defined.
            # Right now, this is just the Calculation class that's skipped with this
            # condition.
            if parent.__base__ != DatabaseTable or not hasattr(parent, "_from_toolkit"):
                continue

            # Grab the input arguments for the _from_toolkit method. This will
            # give a list back like... ['cls', 'structure', 'as_dict']. We
            # don't consider as_dict or cls, so we remove those too
            inputs = inspect.getfullargspec(parent._from_toolkit).args
            inputs.remove("as_dict")
            inputs.remove("cls")

            # Now inputs is current of list of keys that we need. Next, we go
            # through our kwargs and see if we have any of these keys, and if so,
            # grab them.
            matching_inputs = {
                key: kwargs[key] for key in inputs if key in kwargs.keys()
            }

            # We now pass these inputs to the _from_toolkit. If we're missing
            # a required input, this will raise an error here. We only
            # want the expanded input, so we request a dictionary, not object.
            data = parent._from_toolkit(**matching_inputs, as_dict=True)

            # Now add this mixin's data to our collective dictionary
            all_data.update(data)

        # If as_dict is false, we build this into an Object. Otherwise, just
        # return the dictionary
        return all_data if as_dict else cls(**all_data)

    # -------------------------------------------------------------------------
    # Methods creating new archives
    # -------------------------------------------------------------------------

    @classmethod
    def to_archive(cls, filename: Path | str = None):
        """
        Writes the entire database table to an archive file. If you prefer
        a subset of entries for the archive, use the to_archive method
        on your SearchResults instead (e.g. MyTable.objects.filter(...).to_archive())
        """
        cls.objects.all().to_archive(filename)

    @classmethod
    @property
    def archive_fieldset(cls) -> list[str]:

        all_fields = ["id", "updated_at", "created_at", "source"]

        # If calling this method on the base class, just return the sole mix-in.
        if cls == DatabaseTable:
            return all_fields

        # Otherwise we need to go through the mix-ins and add their fields to
        # the list
        all_fields += [
            field for mixin in cls.get_mixins() for field in mixin.archive_fields
        ]

        # Sometimes a column will be disabled by adding "--" in front of the
        # column name. For example, "--band_gap" would exclude storing the band
        # gap in the archive. We look for any columns that start with this
        # and then remove them
        for field in cls.archive_fields:
            if field.startswith("--"):
                all_fields.remove(field.removeprefix("--"))
            else:
                all_fields.append(field)
        return all_fields

    # -------------------------------------------------------------------------
    # Methods that handle loading results from archives
    # -------------------------------------------------------------------------

    # @transaction.atomic  # We can't have an atomic transaction if we use Dask
    @classmethod
    def load_archive(
        cls,
        filename: str | Path = None,
        delete_on_completion: bool = False,
        confirm_override: bool = False,
        parallel: bool = False,
        confirm_sqlite_parallel: bool = False,
    ):
        """
        Reads a compressed zip file made by `objects.to_archive` and loads the data
        back into the Simmate database.

        Typically, users won't call this method directly, but instead use the
        `load_remote_archive` method, which handles downloading the archive
        file from the Simmate website for you.

        #### Parameters

        - `filename`:
            The filename to write the zip file to. By defualt, None will try to
            find a file named "MyExampleTableName-2022-01-25.zip", where the date
            corresponds to version/timestamp. If multiple files match this format
            the most recent date will be used.

        - `delete_on_completion`:
            Whether to delete the archive file once all data is loaded into the
            database. Defaults to False

        - `confirm_override`:
            If the table already has data in it, the user must take particular
            care to downloading new data. This flag makes sure the user has
            made the proper checks to run this action. Default is False.

        - `parallel`:
            Whether to load the data in parallel. If true, this will start
            a local Dask cluster and each data row will be submitted as a task
            to the cluster. This provides substansial speed-ups for loading
            large datasets into the dataset. Default is False.

        - `confirm_sqlite_parallel`:
            If the database backend is sqlite, this parameter ensures the user
            knows what they are doing and know the risks of parallelization.
            Default is False.
        """

        # We disable warnings while loading archives because pymatgen prints
        # a lot of them (for things like rounding or electronegativity alerts).
        warnings.filterwarnings("ignore")

        # make sure the user actually wants to do this!
        cls._confirm_override(
            confirm_override,
            parallel,
            confirm_sqlite_parallel,
        )

        from simmate.toolkit import Structure as ToolkitStructure

        # generate the file name if one wasn't given
        if not filename:
            # The name will be something like "MyExampleTable-2022-01-25.zip".
            # We go through all files that match "MyExampleTable-*.zip" and then
            # grab the most recent date.
            matching_files = [
                file
                for file in Path.cwd().iterdir()
                if file.name.startswith(cls.table_name) and file.suffix == ".zip"
            ]
            # make sure there is at least one file
            if not matching_files:
                raise FileNotFoundError(
                    f"No file found matching the {cls.table_name}-*.zip format"
                )
            # sort the files by date and grab the first
            matching_files.sort(reverse=True)
            filename = matching_files[0]

        # Turn the filename into the full path -- which makes a number of
        # manipulations easier below.
        filename = Path(filename).absolute()

        # uncompress the zip file to the same directory that it is located in
        shutil.unpack_archive(
            filename,
            extract_dir=filename.parent,
        )

        # We will now have a csv file of the same name, which we load into
        # a pandas dataframe
        csv_filename = filename.with_suffix(".csv")  # was ".zip"
        df = pandas.read_csv(csv_filename)

        # BUG: NaN values throw errors when read into SQL databases, so we
        # convert all NaN entries to None. This hacky line was taken from
        #   https://stackoverflow.com/questions/39279824/
        df = df.astype(object).where(df.notna(), None)

        # convert the dataframe to a list of dictionaries that we will iterate
        # through.
        entries = df.to_dict(orient="records")

        # to enable parallelization, we define a function to load a single
        # entry (or row) of data. This allows us to submit the function to Dask.
        def load_single_entry(entry):
            """
            Quick utility function that load a single entry to the database.
            We have this as a internal function in order to allow submitting
            this function to Dask (for parallelization).
            """

            # For all entries, convert the structure_string to a toolkit structure
            if "structure_string" in entry:
                structure_str = entry.pop("structure_string")
                structure = ToolkitStructure.from_database_string(structure_str)
                entry["structure"] = structure
            # OPTIMIZE: is there a better way to do decide which entries need to be
            # converted to toolkit objects?

            # BUG: some columns don't properly convert to python objects, but
            # it seems inconsistent when this is done... For now I just manually
            # convert JSON columns
            json_parsing_columns = ["site_forces", "lattice_stress"]
            for column in json_parsing_columns:
                if column in entry:
                    if entry[column]:  # sometimes it has a value of None
                        entry[column] = json.loads(entry[column])
            # OPTIMIZE: consider applying this to the df column for faster loading

            entry_db = cls.from_toolkit(**entry)
            entry_db.save()

        # now iterate through all entries to save them to the database
        if not parallel:
            # If user doesn't want parallelization, we run these in the main
            # thread and monitor progress
            for entry in track(entries):
                load_single_entry(entry)
        # otherwise we use dask to submit these in batches!
        else:

            from simmate.configuration.dask import batch_submit

            batch_submit(
                function=load_single_entry,
                args_list=entries,
                batch_size=15000,
            )

        # We can now delete the files. The zip file is only deleted if requested.
        csv_filename.unlink()
        if delete_on_completion:
            filename.unlink()  # the zip archive

    @classmethod
    def load_remote_archive(
        cls,
        remote_archive_link: str = None,
        confirm_override: bool = False,
        parallel: bool = False,
        confirm_sqlite_parallel: bool = False,
    ):
        """
        Downloads a compressed zip file made by `objects.to_archive` and loads
        the data back into the Simmate database.

        This method should only be called once -- when you have a completely
        empty database. After this call, all data will be stored locally and
        you don't need to call this method again (even accross python sessions).

        #### Parameters

        - `remote_archive_link`:
            The URL for that the archive will be downloaded from. If not supplied,
            it will default to the table's remote_archive_link attribute.

        - `confirm_override`:
            If the table already has data in it, the user must take particular
            care to downloading new data. This flag makes sure the user has
            made the proper checks to run this action.

        - `parallel`:
            Whether to load the data in parallel. If true, this will start
            a local Dask cluster and each data row will be submitted as a task
            to the cluster. This provides substansial speed-ups for loading
            large datasets into the dataset. Default is False.

        - `confirm_sqlite_parallel`:
            If the database backend is sqlite, this parameter ensures the user
            knows what they are doing and know the risks of parallelization.
            Default is False.
        """

        # make sure the user actually wants to do this!
        cls._confirm_override(
            confirm_override,
            parallel,
            confirm_sqlite_parallel,
        )

        # confirm that we have a link to download from
        if not remote_archive_link:
            # if no link was given we take the input value from class attribute
            if not cls.remote_archive_link:
                raise Exception(
                    "This table does not have a default link to load the archive "
                    " from. You must provide a remote_archive_link."
                )
            remote_archive_link = cls.remote_archive_link

        # tell the user where the data comes from
        if cls._meta.app_label == "third_parties":
            logging.warning(
                "this data is NOT from the Simmate team, so be sure "
                "to visit the provider's website and to cite their work."
                f" This data is from {cls.source} and the following paper "
                f"should be cited: {cls.source_doi}"
            )

        # Predetermine the file name, which is just the ending of the URL
        archive_filename = remote_archive_link.split("/")[-1]

        # Download the archive zip file from the URL to the current working dir
        logging.info("Downloading archive file...")
        urllib.request.urlretrieve(remote_archive_link, archive_filename)
        logging.info("Done.")

        # now that the archive is downloaded, we can load it into our db
        logging.info("Loading data into Simmate database")
        cls.load_archive(
            archive_filename,
            delete_on_completion=True,
            confirm_override=True,  # we already confirmed this above
            parallel=parallel,
            confirm_sqlite_parallel=True,  # we already confirmed this above
        )
        logging.info("Done.")

    @classmethod
    def _confirm_override(
        cls,
        confirm_override: bool,
        parallel: bool,
        confirm_sqlite_parallel: bool,
    ):
        """
        A utility to make sure the user wants to load new data into their table
        and (if they are using sqlite) that they are aware of the risks of
        parallelizing their loading.

        This utility should not be called directly, as it is used within
        load_archive and load_remote_archive.
        """
        # first check if the table has data in it already. We raise errors
        # to stop the user from doing unneccessary and potentiall destructive
        # downloads
        if cls.objects.exists() and not confirm_override:
            # if the user has a third-party app, we can be more specific with
            # our error message.
            if cls._meta.app_label == "third_parties":
                raise Exception(
                    "It looks like you're using a third-party database table and "
                    "that the table already has data in it! This means you already "
                    "called load_remote_archive and don't need to do it again. "
                    "If you are trying reload a newer version of this data, make "
                    "sure you empty this table first. This can be done by "
                    "reseting your database or manually deleting all objects "
                    "with `ExampleTable.objects.all().delete()`"
                )

            # otherwise warning the user of overwriting data with matching
            # primary keys -- and ask them to use confirm_override.
            raise Exception(
                "It looks like this table already has data in it! By loading an "
                "archive, you could potentially overwrite this data. The most "
                "common mistake is non-unique primary keys between your current "
                "data and the archive -- if there is a duplicate primary key, it "
                "will overwrite your data. If you are confident the data is safe "
                "to load into your database, run this command again with "
                "confirm_override=True."
            )

        # Django and Dask can only handle so much for the parallelization
        # of database writing with SQLite. So if the user has SQLite as their
        # backend, we need to stop them from using this feature.
        from simmate.configuration.django.settings import DATABASES

        if parallel and not confirm_sqlite_parallel and "sqlite3" in str(DATABASES):
            raise Exception(
                "It looks like you are trying to run things in parallel but are "
                "using the default database backend (sqlite3), which is not "
                "always stable for massively parallel methods. You can still "
                "do this, but this message serves as a word of caution. "
                "You If you see error messages pop up saying 'database is "
                "locked', then your database is not stable at the rate you're "
                "trying to write data. This is a sign that you should either "
                "(1) switch to a different database backend such as Postgres "
                "or (2) reduce the parallelization of your tasks. If you are "
                "comfortable with these warnings and know what you're doing, "
                "set confirm_sqlite_parallel=True."
            )

    # -------------------------------------------------------------------------
    # Methods that set up the REST API and filters that can be queried with
    # -------------------------------------------------------------------------

    @classmethod
    @property
    @cache
    def api_filterset(cls) -> django_api_filters.FilterSet:
        """
        Dynamically creates a Django Filter from a Simmate database table.

        For example, this function would take
        `simmate.database.third_parties.MatprojStructure`
        and automatically make the following filter:

        ``` python
        from simmate.website.core_components.filters import (
            DatabaseTableFilter,
            Structure,
            Thermodynamics,
        )


        class MatprojStrucureFilter(
            DatabaseTableFilter,
            Structure,
            Thermodynamics,
        ):
            class Meta:
                model = MatprojStructure  # this is database table
                fields = {...} # this combines the fields from Structure/Thermo mixins

            # These attributes are set using the declared filters from
            # the Structure/Thermo mixins
            declared_filter1 = ...
            declared_filter1 = ...
        ```
        """

        # ---------------------------------------------------------------------
        # This is the filter mix-in for the base DatabaseTable class, which
        # ALL other filters will use. We set this up manually.
        # TODO: Consider moving this outside of this method.
        class ApiFilter(django_api_filters.FilterSet):
            class Meta:
                table = DatabaseTable
                fields = dict(
                    # id=["exact"],  --> automatically added...?
                    created_at=["range"],
                    updated_at=["range"],
                )

            def skip_filter(self, queryset, name, value):
                """
                For filter fields that use this method, nothing is done to queryset. This
                is because the filter is typically used within another field. For example,
                the `include_subsystems` field is not applied to the queryset, but it
                is used within the `filter_chemical_system` method.
                """
                return queryset

        # ---------------------------------------------------------------------

        # If calling this method on the base class, just return the sole mix-in.
        if cls == DatabaseTable:
            return ApiFilter

        # First we need to grab the parent mixin classes of the table. For example,
        # the MatprojStructure uses the database mixins ['Structure', 'Thermodynamics']
        # while MatprojStaticEnergy uses ["StaticEnergy"].
        table_mixins = cls.get_mixins()

        # When we make this filter, we want it to inherit from the filters of
        # all mixins available. We therefore grab all these filters into a list.
        # We also add the standard ModelForm class from django-filters.
        filter_mixins = [ApiFilter]
        filter_mixins += [mixin.api_filterset for mixin in table_mixins]

        # "Declared Filters" are ones normally set as an attribute when creating
        # a Django filter, whereas normal filters are provided in a meta. For
        # example...
        #
        # class Structure(filters.FilterSet):
        #     class Meta:
        #         model = StructureTable
        #         fields = dict(
        #             nsites=["range"],  <-------------- field filter
        #             nelements=["range"],  <----------- field filter
        #         )
        #     include_subsystems = filters.BooleanFilter(
        #         field_name="include_subsystems",
        #         label="Include chemical subsystems in results?",
        #         method="skip_filter",
        #     ) <-------------------------------------- declared filter
        #     chemical_system = filters.CharFilter(
        #         method="filter_chemical_system"
        #     ) <-------------------------------------- declared filter
        #
        # I need to separate these out in order to create our class properly
        field_filters = {}
        declared_filters = {}
        filter_methods = {}
        for field, conditions in cls.api_filters.items():
            if isinstance(conditions, django_api_filters.Filter):
                declared_filters.update({field: conditions})
                # For the declared filters, there is sometimes a method that
                # needs to be called. This will be attached to the original
                # mixin as a method of the same name. Here we check if there
                # is a method and attach it. For an example of this, see the
                # "filter_chemical_system" method of the Structure class.
                if conditions.method and conditions.method != "skip_filter":
                    filter_methods.update(
                        {conditions.method: getattr(cls, conditions.method)}
                    )
            else:
                field_filters.update({field: conditions})

        # combine the fields of each filter mixin. Note we use .get_fields instead
        # of .fields to ensure we get a dictionary back
        field_filters.update(
            {
                field: conditions
                for mixin in filter_mixins
                for field, conditions in mixin.get_fields().items()
            }
        )

        # Also combine all declared filters from each mixin
        declared_filters.update(
            {
                name: filter_obj
                for mixin in filter_mixins
                for name, filter_obj in mixin.declared_filters.items()
            }
        )

        # The FilterSet class requires that we set extra fields in the Meta class.
        # We define those here. Note, we accept all fields by default and the
        # excluded fields are only those defined by the supported form_mixins -
        # and we skip the first mixin (which is always DatabaseTableFilter)
        class Meta:
            model = cls
            fields = field_filters

        extra_attributes = {"Meta": Meta, **declared_filters, **filter_methods}
        # BUG: __module__ may need to be set in the future, but we never import
        # these forms elsewhere, so there's no need to set it now.

        # Now we dynamically create a new form class that we can return.
        NewClass = type(cls.table_name, tuple(filter_mixins), extra_attributes)

        # Match the filter name to the table name
        NewClass.filter_name = cls.table_name

        return NewClass

    @classmethod
    @property
    @cache
    def api_filters_extra(cls):
        """
        Finds all columns that aren't covered by the supported Form mix-ins.

        For example, a form made from the database table...

        ``` python
        from simmate.database.base_data_types import (
            table_column,
            Structure,
            Forces,
        )

        class ExampleTable(Structure, Forces):
            custom_column1 = table_column.FloatField()
            custom_column2 = table_column.FloatField()
        ```

        ... would return ...

        ``` python
        ["custom_column1", "custom_column2"]
        ```
        """

        all_columns = cls.api_filterset.base_filters.keys()
        columns_w_mixin = [
            column
            for mixin in cls.get_mixins()
            for column in mixin.api_filterset.base_filters.keys()
        ]
        extra_columns = [
            column
            for column in all_columns
            if column not in columns_w_mixin and column != "id"
        ]
        return extra_columns



# -*- coding: utf-8 -*-

"""
This module provides a single endpoint where all registered workflows can be
accessed for user-convenience. 

This is best for accessing workflows in a quick debugging manner. If you need
to access all workflows, we recommend using this utility instead:
``` python
simmate.workflows.utilities import get_all_workflows 
```
"""

# note, everything starts with an underscore because I don't want these variables
# to be visable -- only the workflow classes.
# The code below is modified from...
#   https://stackoverflow.com/questions/30986470/

from simmate.workflows.utilities import get_all_workflows as _get_all

_workflows = _get_all()
_module = globals()

for _workflow in _workflows:
    _module[_workflow.__name__] = _workflow



# -*- coding: utf-8 -*-

from simmate.calculators.vasp.inputs.potcar_mappings import (
    PBE_ELEMENT_MAPPINGS_LOW_QUALITY,
)
from simmate.calculators.vasp.workflows.base import VaspWorkflow


class Relaxation__Vasp__Quality04(VaspWorkflow):
    """
    Runs a rough VASP geometry optimization.

    `Quality 04` indicates that on a scale from 00 to 04, these are ranked 04 in
    quality (with 04 being the highest quality).

    Note, even though this is currently our highest quality preset, these
    settings are still only suitable for high-throughput calculations or massive
    supercells. Settings are still below MIT and Materials Project quality.

    Typically, you'd only want to run this relaxation on structures that were
    randomly created (and thus likely very unreasonable). Therefore, instead of
    using this calculation, we recommend only using the relaxation/staged
    workflow, which uses this calculation as a fifth step (and final relaxation).
    """

    description_doc_short = "for randomly-created structures"

    # This uses the PBE functional with POTCARs that have lower electron counts
    # and convergence criteria when available.
    functional = "PBE"
    potcar_mappings = PBE_ELEMENT_MAPPINGS_LOW_QUALITY

    # Make the unitcell relatively cubic before relaxing
    standardize_structure = "primitive-LLL"
    symmetry_tolerance = 0.1

    # These are all input settings for this task.
    incar = dict(
        # These settings are the same for all structures regardless of composition.
        PREC="Normal",
        EDIFF=1e-5,
        ENCUT=450,  # !!! Should this be based on the element type?
        ISIF=3,
        NSW=100,
        IBRION=1,
        POTIM=0.02,
        LCHARG=False,
        LWAVE=False,
        KSPACING=0.4,
        # The type of smearing we use depends on if we have a metal, semiconductor,
        # or insulator. So we need to decide this using a keyword modifier.
        multiple_keywords__smart_ismear={
            "metal": dict(
                ISMEAR=1,
                SIGMA=0.06,
            ),
            "non-metal": dict(
                ISMEAR=0,
                SIGMA=0.05,
            ),
        },
    )



# -*- coding: utf-8 -*-

from pathlib import Path

from pymatgen.analysis.transition_state import NEBAnalysis
from pymatgen.core.sites import PeriodicSite
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

from simmate.calculators.vasp.outputs import Vasprun
from simmate.database.base_data_types import Calculation, Structure, table_column
from simmate.toolkit import Structure as ToolkitStructure
from simmate.toolkit.diffusion import MigrationHop as ToolkitMigrationHop
from simmate.toolkit.diffusion import MigrationImages
from simmate.visualization.plotting import MatplotlibFigure


class DiffusionAnalysis(Structure, Calculation):
    class Meta:
        app_label = "workflows"

    api_filters = dict(
        migrating_specie=["exact"],
        vacancy_mode=["exact"],
        atomic_fraction=["range"],
        barrier_cell=["range"],
        npaths_involved=["range"],
    )

    migrating_specie = table_column.CharField(max_length=4, blank=True, null=True)
    """
    The element of the diffusion atom (e.g. "Li")
    """

    vacancy_mode = table_column.BooleanField(blank=True, null=True)
    """
    (if relevent) Whether vacancy or interstitial diffusion was used.
    """

    atomic_fraction = table_column.FloatField(blank=True, null=True)
    """
    Atomic fraction of the diffusion ion in the bulk structure.
    """

    # NOTE: barrier_cell, paths_involved, and npaths_involved need to be updated
    # by evaluating all MigrationHops.

    barrier_cell = table_column.FloatField(blank=True, null=True)
    """
    The energy barrier corresponding to the lowest-barrier percolation network
    """

    paths_involved = table_column.CharField(max_length=100, blank=True, null=True)
    """
    The symmetrically-unqiue pathways involved in the lowest-barrier 
    percolation network. This is given as a list of IDs.
    """
    # TODO: consider making this a relationship to MigrationPathways

    npaths_involved = table_column.IntegerField(blank=True, null=True)
    """
    The number of symmetrically-unqiue pathways involved in the lowest-barrier 
    percolation network
    """

    @classmethod
    def from_toolkit(
        cls,
        migrating_specie: str = None,
        vacancy_mode: bool = None,
        as_dict: bool = False,
        **kwargs,
    ):
        # the algorithm doesn't change for this method, but we do want to add
        # a few extra columns. Therefore we make the dictionary as normal and
        # then add those extra columns here.
        structure_dict = super().from_toolkit(as_dict=True, **kwargs)

        # now add the two extra columns
        structure_dict["migrating_specie"] = migrating_specie
        # Structure should be present -- otherwise an error would have been
        # raised by the call to super()
        structure_dict["atomic_fraction"] = (
            kwargs["structure"].composition.get_atomic_fraction(migrating_specie)
            if migrating_specie
            else None
        )

        # If as_dict is false, we build this into an Object. Otherwise, just
        # return the dictionary
        return structure_dict if as_dict else cls(**structure_dict)

    @classmethod
    def from_directory(cls, directory: Path):
        """
        Creates a new database entry from a directory that holds diffusion analysis
        results. For now, this assumes the directory holds vasp output files.
        """

        # I assume the directory is from a vasp calculation, but I need to update
        # this when I begin adding new calculators.

        # For now, I only grab the structure from the static-energy and store
        # it in the DiffusionAnalysis table.
        found_dir = False
        for name in directory.iterdir():
            if name.is_dir() and name.stem.startswith("static-energy"):
                static_energy_dir = name
                found_dir = True
                break
        if not found_dir:
            raise Exception(
                "Unable to detect 'static-energy' directory and therefore unable "
                "to determine the bulk structure used for this analysis."
            )

        bulk_filename = static_energy_dir / "POSCAR"
        bulk_structure = ToolkitStructure.from_file(bulk_filename)
        # BUG: I assume the directory location but this will fail if changed.

        # Save a diffusion analysis object so we can connect all other data
        # to it.
        analysis_db = cls.from_toolkit(
            structure=bulk_structure,
            vacancy_mode=True,  # assume this for now
        )
        analysis_db.save()

        # load the remaining data - such as migration hops
        analysis_db.update_from_directory(directory)

        return analysis_db

    def update_from_directory(self, directory: Path):
        # NOTE: This method is not called at the end of the workflow as
        # subflows often created the data already.

        # Iterate through all the subdirectories that start with "migration_hop*".
        # We also need to make sure we only grab directories because there are
        # also cifs present that match this naming convention.
        # We ignore the number when saving to avoid overwriting data.
        migration_directories = [
            f.absolute()
            for f in directory.iterdir()
            if f.is_dir() and "single-path" in f.stem
        ]

        # now save each migration hop present
        for migration_dir in migration_directories:
            hop = self.migration_hops.field.model.from_directory(
                directory=migration_dir,
            )
            hop.diffusion_analysis_id = self.id
            hop.save()


class MigrationHop(Calculation):
    class Meta:
        app_label = "workflows"

    archive_fields = [
        "site_start",
        "site_end",
        "index_start",
        "index_end",
        "number",
    ]

    api_filters = dict(
        site_start=["exact"],
        site_end=["exact"],
        number=["range"],
        length=["range"],
        dimension_path=["range"],
        dimension_host_lattice=["range"],
        energy_barrier=["range"],
    )

    # is_from_hop_obj = table_column.BooleanField(blank=True, null=True)
    # source = hop_obj / endpoints / images

    # OPTIMIZE: site_start and site_end
    # Really, this is a list of float values, but I save it as a string.
    # For robustness, should I save cartesian coordinates and/or lattice as well?
    # Does the max length make sense here and below?
    # I could also store as JSON since it's a list of coords.

    site_start = table_column.CharField(max_length=100, blank=True, null=True)
    """
    The starting fractional coordinates of the diffusing site.
    """

    site_end = table_column.CharField(max_length=100, blank=True, null=True)
    """
    The ending fractional coordinates of the diffusing site.
    """

    # BUG: the init script for MigrationHop can't identify the site indexes
    # properly but they should be the same as before because it is a symmetrized
    # structure. Note that even if I'm wrong in some case -- this will have
    # no effect because iindex and eindex are only used in one portion of
    # the hash as well as for printing the __str__ of the object.

    index_start = table_column.IntegerField(blank=True, null=True)
    """
    The starting site index of the diffusing site.
    """

    index_end = table_column.IntegerField(blank=True, null=True)
    """
    The ending site index of the diffusing site.
    """

    number = table_column.IntegerField(blank=True, null=True)
    """
    The expected index in DistinctPathFinder.get_paths. The shortest path is index 0
    and they are all ordered by increasing length.
    """

    length = table_column.FloatField(blank=True, null=True)
    """
    The length/distance of the pathway from start to end. Note, this is 
    linear measurement -- i.e. is it the length of the linear interpolated path.
    """

    dimension_path = table_column.IntegerField(blank=True, null=True)
    """
    The pathway dimensionality using the Larson scoring parameter.
    """

    dimension_host_lattice = table_column.IntegerField(blank=True, null=True)
    """
    Ignoring all atoms of the diffusing species, this is the dimensionality
    of the remaining atoms of the lattice using the Larson scoring parameter.
    """

    energy_barrier = table_column.FloatField(blank=True, null=True)
    """
    The energy barrier in eV. This evaluates all images of the diffusion pathway
    and find the difference between the maximum and minimum image energies.
    """

    # TODO:
    # Distance of the pathway relative to the shortest pathway distance
    # in the structure using the formula: (D - Dmin)/Dmin
    # distance_rel_min = models.FloatField()

    """ Relationships """
    # Each MigrationHop corresponds to the full analysis of one Structure, which
    # can have many MigrationHop(s)
    diffusion_analysis = table_column.ForeignKey(
        DiffusionAnalysis,
        on_delete=table_column.CASCADE,
        related_name="migration_hops",
        blank=True,
        null=True,
    )

    def write_output_summary(self, directory: Path):
        super().write_output_summary(directory)
        self.write_neb_diagram_plot(directory=directory)
        self.write_migration_images(directory=directory)

    def get_migration_images(self) -> MigrationImages:
        structures = self.migration_images.order_by("number").to_toolkit()
        migration_images = MigrationImages(structures)
        return migration_images

    def write_migration_images(self, directory: Path):
        migration_images = self.get_migration_images()
        structure_sum = migration_images.get_sum_structure()
        structure_sum.to("cif", directory / "simmate_path_relaxed_neb.cif")

    # TODO:
    # image_start --> OneToOneField for specific MigrationHop
    # image_end --> OneToOneField for specific MigrationHop
    # image_transition_state --> OneToOneField for specific MigrationHop

    # Just like Relaxation points to IonicSteps, NEB will point to MigrationImages

    # BUG: because of rounding in the from_toolkit method, the get_sc_structures
    # is unable to identify equivalent sites. I opened an issue for this
    # with their team:
    #   https://github.com/materialsvirtuallab/pymatgen-analysis-diffusion/issues/296
    def to_migration_hop_toolkit(self) -> ToolkitMigrationHop:
        """
        converts the database MigrationHop to a toolkit MigrationHop
        """
        # The bulk crystal structure is stored in the diffusion analysis table
        structure = self.diffusion_analysis.to_toolkit()

        # pathways require a symmetrized structure
        # BUG: I need to ensure the symmetrized structure comes out the same
        # as before, so I need to be careful with updates and hardcoding here.
        sga = SpacegroupAnalyzer(structure, symprec=0.1)
        symm_structure = sga.get_symmetrized_structure()

        isite_new = PeriodicSite(
            species=self.diffusion_analysis.migrating_specie,
            coords=[float(x) for x in self.site_start.split(" ")],
            lattice=symm_structure.lattice,
        )
        esite_new = PeriodicSite(
            species=self.diffusion_analysis.migrating_specie,
            coords=[float(x) for x in self.site_end.split(" ")],
            lattice=symm_structure.lattice,
        )

        path = ToolkitMigrationHop(isite_new, esite_new, symm_structure)
        # BUG: see comment attached to definition of these fields
        path.iindex = self.index_start
        path.eindex = self.index_end

        # if the pathways match, then we can return the pathway object!
        return path

    def to_neb_toolkit(self) -> NEBAnalysis:

        images = self.migration_images.all()
        structures = self.migration_images.to_toolkit()

        neb_toolkit = NEBAnalysis(
            r=[i.structure_distance for i in images],
            energies=[i.energy for i in images],
            forces=[i.force_tangent for i in images],
            structures=structures,
        )
        return neb_toolkit

    def update_from_directory(self, directory: Path):

        # check if we have a VASP directory
        vasprun_filename = directory / "vasprun.xml"
        if not vasprun_filename.exists():
            raise Exception("Only VASP outputs are supported for NEB")

        from simmate.calculators.vasp.outputs import Vasprun

        vasprun = Vasprun.from_directory(directory)
        self.update_from_neb_toolkit(vasprun.neb_results)

    @classmethod
    def from_vasp_run(cls, vasprun: Vasprun):
        return cls.from_neb_toolkit(neb_results=vasprun.neb_results)

    @classmethod
    def from_neb_toolkit(cls, neb_results: NEBAnalysis):

        # This table entry will actually be completely empty... It only
        # serves to link the MigrationImages together. We only create this
        # if we are updating
        hop_db = cls()
        hop_db.save()

        # build out the images in the related table
        hop_db.update_from_neb_toolkit(neb_results)

        return hop_db

    def update_from_neb_toolkit(self, neb_results: NEBAnalysis):

        # build migration images and link them to this parent object.
        # Note, the start/end Migration images will exist already in the
        # relaxation database table. We still want to save them again here for
        # easy access.
        for image_number, image_data in enumerate(
            zip(
                neb_results.structures,
                neb_results.energies,
                neb_results.forces,
                neb_results.r,
            )
        ):
            image, energy, force, distance = image_data
            image_db = self.migration_images.field.model.from_toolkit(
                structure=image,
                number=image_number,
                force_tangent=force,
                energy=energy,
                structure_distance=distance,
                migration_hop_id=self.id,
            )
            image_db.save()

    @classmethod
    def from_toolkit(  # from_migration_hop_toolkit -- registration uses this
        cls,
        migration_hop: ToolkitMigrationHop = None,
        as_dict: bool = False,
        number: int = None,
        **kwargs,
    ):
        # the algorithm doesn't change for this method, but we do want to add
        # a few extra columns. Therefore we make the dictionary as normal and
        # then add those extra columns here.
        structure_dict = super().from_toolkit(as_dict=True, **kwargs)

        if migration_hop:
            # convert the pathway object into the database table format
            hop_dict = dict(
                site_start=" ".join(str(c) for c in migration_hop.isite.frac_coords),
                site_end=" ".join(str(c) for c in migration_hop.esite.frac_coords),
                index_start=migration_hop.iindex,
                index_end=migration_hop.eindex,
                length=migration_hop.length,
                # diffusion_analysis_id=diffusion_analysis_id,
                **kwargs,
            )
        else:
            hop_dict = {}

        all_data = {**structure_dict, **hop_dict}

        # If as_dict is false, we build this into an Object. Otherwise, just
        # return the dictionary
        return all_data if as_dict else cls(**all_data)


class MigrationImage(Structure):
    class Meta:
        app_label = "workflows"

    archive_fields = [
        "number",
        "force_tangent",
        "structure_distance",
        "energy",
    ]

    api_filters = dict(
        number=["exact"],
        force_tangent=["range"],
        energy=["range"],
        structure_distance=["range"],
    )

    number = table_column.IntegerField()
    """
    The image number. Note, the starting image is always 0 and the ending
    image is always -1. The choice of -1 is because we often add the ending
    image to the database table before knowing how many midpoint images we
    are going to create.
    """

    force_tangent = table_column.FloatField(blank=True, null=True)
    """
    The tangent force of image. Not to be confused with forces from the Forces mix-in.
    """

    # Note, we only care about the total energy for NEB images -- not the other
    # fields that the Thermodynamics mix-in provides. This is why we set this
    # field directly, rather than using the mix-in.
    energy = table_column.FloatField(blank=True, null=True)
    """
    The calculated total energy. Units are in eV. 
    """

    structure_distance = table_column.FloatField(blank=True, null=True)
    """
    The fingerprint distance of the image from the starting image. A smaller
    means the structures are more similar, with 0 being an exact match.
    """

    # We don't need the source column for the MigrationImage class because we
    # instead stored the source on the DiffusionAnalysis object. This line
    # deletes the source column from our Structure mix-in.
    source = None

    migration_hop = table_column.ForeignKey(
        MigrationHop,
        on_delete=table_column.CASCADE,
        related_name="migration_images",
    )


class NebDiagram(MatplotlibFigure):
    def get_plot(results: MigrationHop):
        neb_results = results.to_neb_toolkit()
        plot = neb_results.get_plot()
        return plot


# register all plotting methods to the database table
for _plot in [NebDiagram]:
    _plot.register_to_class(MigrationHop)



# -*- coding: utf-8 -*-

"""

> :warning: This file is only for use by the Simmate team. Users should instead
access data via the load_remote_archive method.

This file is for pulling Materials Project data into the Simmate database. 
PyMatGen offers an easy way to do this in python -- the MPRester class. All you
need is [an API key from their site](https://materialsproject.org/open) and pymatgen
installed. For now, we only pull the mp-id, structure, and final energy.

"""

from django.db import transaction
from rich.progress import track

from simmate.database.third_parties import MatprojStructure


@transaction.atomic
def load_all_structures(
    api_key: str,
    update_stabilities: bool = False,
):
    """
    Only use this function if you are part of the Simmate dev team!

    Loads all structures directly for the Material Project database into the
    local Simmate database.

    #### Parameters

    - `api_key`:
        Your Materials Project API key.
    - `criteria`:
        Filtering criteria for which structures to load. The default is all
        existing structures (137,885 as of 2022-01-16), which will take rouhghly
        15 min to complete (not including stabilities).
    - `update_stabilities`:
        Whether to run update_all_stabilities on the database table. Note this
        will add over an hour to this process. Default is True.
    """

    try:
        from mp_api.client import MPRester
    except:
        raise Exception(
            "To use this method, MP-API is required. Please install it "
            "with `pip install mp-api"
        )

    # Connect to their database with personal API key
    with MPRester(api_key) as mpr:

        # For the filtered structures, this lists off which properties to grab.
        # All possible properties can be listed with:
        #   mpr.summary.available_fields
        fields_to_load = [
            # "builder_meta",
            # "nsites",
            # "elements",
            # "nelements",
            # "composition",
            # "composition_reduced",
            # "formula_pretty",
            # "formula_anonymous",
            # "chemsys",
            # "volume",
            # "density",
            # "density_atomic",
            # "symmetry",
            # "property_name",
            "material_id",
            # "deprecated",
            # "deprecation_reasons",
            "last_updated",
            # "origins",
            # "warnings",
            "structure",
            # "task_ids",
            "uncorrected_energy_per_atom",
            "energy_per_atom",
            # "formation_energy_per_atom",
            # "energy_above_hull",
            # "is_stable",
            # "equilibrium_reaction_energy_per_atom",
            # "decomposes_to",
            # "xas",
            # "grain_boundaries",
            "band_gap",
            # "cbm",
            # "vbm",
            # "efermi",
            "is_gap_direct",
            # "is_metal",
            # "es_source_calc_id",
            # "bandstructure",
            # "dos",
            # "dos_energy_up",
            # "dos_energy_down",
            "is_magnetic",
            # "ordering",
            "total_magnetization",
            # "total_magnetization_normalized_vol",
            # "total_magnetization_normalized_formula_units",
            # "num_magnetic_sites",
            # "num_unique_magnetic_sites",
            # "types_of_magnetic_species",
            # "k_voigt",
            # "k_reuss",
            # "k_vrh",
            # "g_voigt",
            # "g_reuss",
            # "g_vrh",
            # "universal_anisotropy",
            # "homogeneous_poisson",
            # "e_total",
            # "e_ionic",
            # "e_electronic",
            # "n",
            # "e_ij_max",
            # "weighted_surface_energy_EV_PER_ANG2",
            # "weighted_surface_energy",
            # "weighted_work_function",
            # "surface_anisotropy",
            # "shape_factor",
            # "has_reconstructed",
            # "possible_species",
            # "has_props",
            "theoretical",
        ]

        # now make the query and grab everything from the Materials Project!
        # the output dictionary is given back within a list, where each entry is
        # a specific structure (so a single mp-id)
        # Note: this is a very large query, so make sure your computer has enough
        # memory (RAM >10GB) and a stable internet connection.
        # data = mpr.summary.search(
        #     all_fields=False,
        #     fields=fields_to_load,
        #     deprecated=False,
        #     # !!! DEV NOTE: you can uncomment these lines for quick testing
        #     # num_chunks=3,
        #     chunk_size=100,
        # )

        # BUG: The search above is super unstable, so instead, I grab all mp-id
        # in one search, then make individual queries for the data of each
        # after that.
        # This takes about 30 minutes.
        mp_ids = mpr.summary.search(
            all_fields=False,
            fields=["material_id"],
            deprecated=False,
        )
        data = []
        for entry in track(mp_ids):
            result = mpr.summary.search(
                material_ids=[entry.material_id],
                all_fields=False,
                fields=fields_to_load,
            )
            data.append(result[0])

    # return data

    # Let's iterate through each structure and save it to the database
    # This also takes a while, so we use a progress bar
    failed_entries = []
    for entry in track(data):
        try:
            # convert the data to a Simmate database object
            structure_db = MatprojStructure.from_toolkit(
                id=str(entry.material_id),
                structure=entry.structure,
                energy=entry.energy_per_atom * entry.structure.num_sites,
                energy_uncorrected=entry.uncorrected_energy_per_atom
                * entry.structure.num_sites,
                updated_at=fix_timezone(entry.last_updated),
                band_gap=entry.band_gap,
                is_gap_direct=entry.is_gap_direct,
                is_magnetic=entry.is_magnetic,
                total_magnetization=entry.total_magnetization,
                is_theoretical=entry.theoretical,
            )

            # and save it to our database!
            structure_db.save()
        except:
            failed_entries.append(entry)

    return failed_entries

    # # once all structures are saved, let's update the Thermodynamic columns
    # if update_stabilities:
    #     MatprojStructure.update_all_stabilities()


# This fix function below is a copy/paste from...
#   https://stackoverflow.com/questions/18622007/

from django.conf import settings
from django.utils.timezone import make_aware


def fix_timezone(naive_datetime):

    settings.TIME_ZONE  # 'UTC'
    aware_datetime = make_aware(naive_datetime)
    aware_datetime.tzinfo  # <UTC>

    return aware_datetime
